/**
 * Copyright © 2014-2021 The SiteWhere Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.sitewhere.instance.microservice;

import com.sitewhere.instance.configuration.InstanceManagementTenantConfiguration;
import com.sitewhere.instance.configuration.InstanceManagementTenantEngineModule;
import com.sitewhere.instance.pipeline.PipelineLogMonitor;
import com.sitewhere.instance.spi.microservice.IInstanceManagementTenantEngine;
import com.sitewhere.instance.spi.pipeline.IPipelineLogMonitor;
import com.sitewhere.microservice.lifecycle.CompositeLifecycleStep;
import com.sitewhere.microservice.multitenant.MicroserviceTenantEngine;
import com.sitewhere.spi.SiteWhereException;
import com.sitewhere.spi.microservice.lifecycle.ICompositeLifecycleStep;
import com.sitewhere.spi.microservice.lifecycle.ILifecycleProgressMonitor;
import com.sitewhere.spi.microservice.multitenant.IMicroserviceTenantEngine;
import com.sitewhere.spi.microservice.multitenant.ITenantEngineModule;

import io.sitewhere.k8s.crd.tenant.engine.SiteWhereTenantEngine;

/**
 * Implementation of {@link IMicroserviceTenantEngine} that implements instance
 * management functionality.
 */
public class InstanceManagementTenantEngine extends MicroserviceTenantEngine<InstanceManagementTenantConfiguration>
	implements IInstanceManagementTenantEngine {

    /** Pipeline log monitor */
    private IPipelineLogMonitor pipelineLogMonitor = new PipelineLogMonitor();

    public InstanceManagementTenantEngine(SiteWhereTenantEngine tenantEngineResource) {
	super(tenantEngineResource);
    }

    /*
     * @see com.sitewhere.spi.microservice.multitenant.IMicroserviceTenantEngine#
     * getConfigurationClass()
     */
    @Override
    public Class<InstanceManagementTenantConfiguration> getConfigurationClass() {
	return InstanceManagementTenantConfiguration.class;
    }

    /*
     * @see com.sitewhere.spi.microservice.multitenant.IMicroserviceTenantEngine#
     * createConfigurationModule()
     */
    @Override
    public ITenantEngineModule<InstanceManagementTenantConfiguration> createConfigurationModule() {
	return new InstanceManagementTenantEngineModule(this, getActiveConfiguration());
    }

    /*
     * @see com.sitewhere.spi.microservice.multitenant.IMicroserviceTenantEngine#
     * tenantInitialize(com.sitewhere.spi.microservice.lifecycle.
     * ILifecycleProgressMonitor)
     */
    @Override
    public void tenantInitialize(ILifecycleProgressMonitor monitor) throws SiteWhereException {
	// Create step that will initialize components.
	ICompositeLifecycleStep init = new CompositeLifecycleStep("Initialize " + getComponentName());

	// Initialize pipeline log monitor.
	init.addInitializeStep(this, getPipelineLogMonitor(), true);

	// Execute initialization steps.
	init.execute(monitor);
    }

    /*
     * @see com.sitewhere.spi.microservice.multitenant.IMicroserviceTenantEngine#
     * tenantStart(com.sitewhere.spi.microservice.lifecycle.
     * ILifecycleProgressMonitor)
     */
    @Override
    public void tenantStart(ILifecycleProgressMonitor monitor) throws SiteWhereException {
	// Create step that will start components.
	ICompositeLifecycleStep start = new CompositeLifecycleStep("Start " + getComponentName());

	// Start pipeline log monitor.
	start.addStartStep(this, getPipelineLogMonitor(), true);

	// Execute startup steps.
	start.execute(monitor);
    }

    /*
     * @see com.sitewhere.spi.microservice.multitenant.IMicroserviceTenantEngine#
     * tenantStop(com.sitewhere.spi.microservice.lifecycle.
     * ILifecycleProgressMonitor)
     */
    @Override
    public void tenantStop(ILifecycleProgressMonitor monitor) throws SiteWhereException {
	// Create step that will stop components.
	ICompositeLifecycleStep stop = new CompositeLifecycleStep("Stop " + getComponentName());

	// Stop pipeline log monitor.
	stop.addStopStep(this, getPipelineLogMonitor());

	// Execute shutdown steps.
	stop.execute(monitor);
    }

    /*
     * @see com.sitewhere.instance.spi.microservice.IInstanceManagementTenantEngine#
     * getPipelineLogMonitor()
     */
    @Override
    public IPipelineLogMonitor getPipelineLogMonitor() {
	return pipelineLogMonitor;
    }
}
